#include <assert.h>
#include "connect4.h"
#include "connect4_stretches.h"

/* initialize board for the start of a new game; WHITE begins */
void connect4_init(connect4_board* board) {
   *board = (connect4_board) {0};
}

/* return true, if the game is finished */
bool connect4_finished(connect4_board* board) {
   return board->finished;
}

/* return true, if the game is finished and nobody has won */
bool connect4_tied(connect4_board* board) {
   return board->finished && board->winner == CONNECT4_FREE;
}

/* return CONNECT4_WHITE, if white has won, or
          CONNECT4_BLACK, if black has won */
enum connect4_slot connect4_winner(connect4_board* board) {
   return board->winner;
}

/* return the next player, i.e. CONNECT4_WHITE or CONNECT4_BLACK
   PRE: the game is not yet finished */
enum connect4_slot connect4_next(connect4_board* board) {
   assert(!board->finished);
   if (board->filled % 2 == 0) {
      return CONNECT4_WHITE;
   } else {
      return CONNECT4_BLACK;
   }
}

/* return true if this move is valid */
bool connect4_valid_move(connect4_board* board, connect4_move move) {
   if (board->finished) return false;
   if (move < 1 || move > COLUMNS) return false;
   if (board->height[move-1] == ROWS) return false;
   return true;
}

/* check if the game is finished */
void static check_status(connect4_board* board) {
   for (int si = 0; si < sizeof stretch / sizeof(struct stretch); ++si) {
      int i = stretch[si].i; int j = stretch[si].j;
      int di = stretch[si].di; int dj = stretch[si].dj;
      enum connect4_slot last_token = CONNECT4_FREE;
      unsigned char len = 0;
      for (int index = 0; index < stretch[si].len; ++index) {
	 enum connect4_slot token = board->board[i][j];
	 if (token == last_token) {
	    ++len;
	 } else {
	    len = 1;
	    last_token = token;
	 }
	 if (len >= WINLEN && token != CONNECT4_FREE) {
	    board->finished = true;
	    board->winner = token;
	    return;
	 }
	 i += di; j += dj;
      }
   }
   if (board->filled == ROWS * COLUMNS) {
      board->finished = true;
      board->winner = CONNECT4_FREE;
   }
}

/* perform the given move on the given board
   PRE: this move must be valid */
void connect4_perform_move(connect4_board* board, connect4_move move) {
   assert(connect4_valid_move(board, move));
   enum connect4_slot token = connect4_next(board);
   board->board[board->height[move-1]++][move-1] = token;
   ++board->filled;
   check_status(board);
   board->last_move = move;
   ++board->number_of_moves;
}

/* return the set of valid moves */
connect4_moves_set connect4_valid_moves(connect4_board* board) {
   if (board->finished) return 0;
   connect4_moves_set result = 0;
   for (int move = 0; move < COLUMNS; ++move) {
      if (board->height[move] < ROWS) {
	 result |= 1<<move;
      }
   }
   return result;
}

/* return true if the game has already started */
bool connect4_started(connect4_board* board) {
   return board->number_of_moves > 0;
}

/* return last move
   PRE: connect4_started returns true */
connect4_move connect4_last_move(connect4_board* board) {
   assert(board->number_of_moves > 0);
   return board->last_move;
}
