#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <afblib/inbuf.h>
#include <afblib/outbuf.h>
#include "connect4.h"
#include "connect4_client.h"
#include "connect4_packets.h"

void c4_init(struct connect4_service* c4, int fd) {
   c4->fd = fd;
   c4->ibuf = (inbuf) {fd};
   c4->obuf = (outbuf) {fd};
}

bool c4_login(struct connect4_service* c4, char* login) {
   /* send request */
   connect4_request request = {CONNECT4_LOGIN, login};
   if (!gen_request(&c4->obuf, &request)) return false;
   if (!outbuf_flush(&c4->obuf)) return false;
   /* receive response */
   connect4_response response = {0};
   if (!parse_response(&c4->ibuf, &response)) return false;
   if (response.result) free(response.result);
   return response.succeeded;
}

/* comparison function for qsort(3) of a strlist */
int compare_strings(const void* p1, const void* p2) {
   char* s1 = *((char **) p1);
   char* s2 = *((char **) p2);
   return strcmp((char *) s1, (char *) s2);
}

bool c4_users(struct connect4_service* c4, strlist* users) {
   /* send request */
   connect4_request request = {CONNECT4_USERS, 0};
   if (!gen_request(&c4->obuf, &request)) return false;
   if (!outbuf_flush(&c4->obuf)) return false;
   /* receive response */
   connect4_response response = {0};
   if (!parse_response(&c4->ibuf, &response)) return false;
   users->len = 0;
   if (response.succeeded && response.result) {
      char* user = 0;
      bool space = true;
      char* cp = response.result;
      do {
	 if (*cp == ' ' || *cp == 0) {
	    if (!space) {
	       char ch = *cp;
	       *cp = 0; char* userdup = strdup(user); *cp = ch;
	       if (!userdup) return false;
	       if (!strlist_push(users, userdup)) return false;
	       space = true;
	    }
	 } else {
	    if (space) {
	       user = cp;
	       space = false;
	    }
	 }
      } while (*cp++);
      free(response.result);
      /* sort list of users */
      qsort(users->list, users->len, sizeof(char *), compare_strings);
   }
   return response.succeeded;
}

enum game_status {GAME_INIT, GAME_CONTINUE, GAME_TIED, GAME_LOST, GAME_WON};
char* status_keywords[] = {"init", "continue", "tied", "lost", "won"};

bool parse_game_status(char** cp, enum game_status* statusp) {
   char* s = *cp;
   unsigned int len = 0;
   while (**cp != ' ' && **cp != 0) {
      ++*cp; ++len;
   }
   if (len == 0) return false;
   if (**cp) ++*cp; /* skip delimiting space */
   for (enum game_status status = GAME_INIT;
	 status < sizeof(status_keywords) / sizeof(char*);
	 ++status) {
      if (strncmp(status_keywords[status], s, len) == 0) {
	 *statusp = status;
	 return true;
      }
   }
   return false;
}

bool parse_game_opponent(char** cp, char** opponent) {
   char* s = *cp;
   unsigned int len = 0;
   while (**cp != ' ' && **cp != 0) {
      ++*cp; ++len;
   }
   if (len == 0) return false;
   if (**cp) ++*cp; /* skip delimiting space */
   char* opp = malloc(len + 1);
   if (!opp) return false;
   strncpy(opp, s, len + 1);
   *opponent = opp;
   return true;
}

bool parse_game_move(char** cp, connect4_move* movep) {
   char* s = *cp;
   unsigned int len = 0;
   while (isdigit(**cp)) {
      ++*cp; ++len;
   }
   if (len == 0) return false;
   if (**cp == ' ') ++*cp; /* skip delimiting space */
   connect4_move move = atoi(s);
   if (move < 1 || move > COLUMNS) return false;
   *movep = move;
   return true;
}

static bool c4_start_game(struct connect4_service* c4,
      char* player, connect4_board* board, char** opponent) {
   /* send request */
   connect4_request request = {CONNECT4_PLAY, player};
   if (!gen_request(&c4->obuf, &request)) return false;
   if (!outbuf_flush(&c4->obuf)) return false;
   /* receive response */
   connect4_response response = {0};
   if (!parse_response(&c4->ibuf, &response)) return false;
   if (!response.result) return false;
   if (!response.succeeded) {
      free(response.result);
      return false;
   }
   /* parse status message and setup initial board */
   char* cp = response.result;
   enum game_status status;
   connect4_init(board);
   if (!parse_game_status(&cp, &status)) return false;
   if (status == GAME_INIT) {
      if (!parse_game_opponent(&cp, opponent)) return false;
   } else if (status == GAME_CONTINUE) {
      connect4_move move;
      if (!parse_game_move(&cp, &move)) return false;
      if (!parse_game_opponent(&cp, opponent)) return false;
      if (!connect4_valid_move(board, move)) return false;
      connect4_perform_move(board, move);
   } else {
      return false;
   }
   free(response.result);
   return true;
}

bool c4_play(struct connect4_service* c4,
      connect4_board* board, char** opponent) {
   return c4_start_game(c4, 0, board, opponent);
}

bool c4_invite(struct connect4_service* c4, char* player,
      connect4_board* board) {
   char* opponent;
   if (!c4_start_game(c4, player, board, &opponent)) return false;
   free(opponent);
   return true;
}

bool c4_move(struct connect4_service* c4,
      connect4_board* board, connect4_move move) {
   if (!connect4_valid_move(board, move)) return false;
   connect4_perform_move(board, move);
   char move_string[4] = {0};
   snprintf(move_string, sizeof move_string - 1, "%d", move);
   /* send request */
   connect4_request request = {CONNECT4_MOVE, move_string};
   if (!gen_request(&c4->obuf, &request)) return false;
   if (!outbuf_flush(&c4->obuf)) return false;
   /* receive response */
   connect4_response response = {0};
   if (!parse_response(&c4->ibuf, &response)) return false;
   if (!response.result) return false;
   if (!response.succeeded) {
      free(response.result);
      return false;
   }
   char* cp = response.result;
   enum game_status status;
   if (!parse_game_status(&cp, &status) || status == GAME_INIT ||
	 !parse_game_move(&cp, &move)) {
      free(response.result);
      return false;
   }
   if (!connect4_finished(board)) {
      if (!connect4_valid_move(board, move)) return false;
      connect4_perform_move(board, move);
   }
   free(response.result);
   return true;
}

bool c4_quit(struct connect4_service* c4) {
   /* send request */
   connect4_request request = {CONNECT4_QUIT, 0};
   if (!gen_request(&c4->obuf, &request)) return false;
   if (!outbuf_flush(&c4->obuf)) return false;
   /* receive response */
   connect4_response response = {0};
   if (!parse_response(&c4->ibuf, &response)) return false;
   if (response.result) free(response.result);
   return response.succeeded;
}

void c4_hangup(struct connect4_service* c4) {
   close(c4->fd);
   inbuf_free(&c4->ibuf);
   outbuf_free(&c4->obuf);
}
