#include <afblib/strlist.h>
#include <afblib/tokenizer.h>
#include <afblib/inbuf_readline.h>
#include <string.h>
#include <stdbool.h>
#include "connect4_packets.h"

static struct command {
   enum connect4_command token;
   char* name;
} commands[] = {
   {CONNECT4_LOGIN, "login"},
   {CONNECT4_USERS, "users"},
   {CONNECT4_PLAY, "play"},
   {CONNECT4_MOVE, "move"},
   {CONNECT4_QUIT, "quit"},
};

/* parse one request within `s' (without terminating CR LF) and setup
   `request' accordingly; return true in case of success */
bool parse_request(stralloc* s, connect4_request* request) {
   stralloc_0(s); /* required by tokenizer */
   strlist tokens = {0};
   if (!tokenizer(s, &tokens)) return false;
   if (tokens.len == 0) return false;
   if (tokens.len > 2) return false;
   bool found = false;
   int cmd;
   for (cmd = 0; cmd < sizeof commands / sizeof(struct command); ++cmd) {
      if (strcmp(tokens.list[0], commands[cmd].name) == 0) {
	 found = true; break;
      }
   }
   if (found) {
      request->command = commands[cmd].token;
      if (tokens.len == 2) {
	 request->param = strdup(tokens.list[1]);
      } else {
	 request->param = 0;
      }
   }
   strlist_free(&tokens);
   return found;
}

/* generate one response string (without terminating CR LF)
   from `response' into `s' */
bool gen_response(connect4_response* response, stralloc* s) {
   s->len = 0;
   if (response->succeeded) {
      if (!stralloc_cats(s, "OK")) return false;
   } else {
      if (!stralloc_cats(s, "FAILED")) return false;
   }
   if (response->result) {
      if (!stralloc_cats(s, " ")) return false;
      if (!stralloc_cats(s, response->result)) return false;
   }
   return true;
}

/* generate one response string (without terminating CR LF)
   from `succeeded' and `msg' into `s' */
bool gen_response_packet(bool succeeded, stralloc* msg, stralloc* s) {
   s->len = 0;
   if (succeeded) {
      if (!stralloc_cats(s, "OK")) return false;
   } else {
      if (!stralloc_cats(s, "FAILED")) return false;
   }
   if (msg->len > 0) {
      if (!stralloc_cats(s, " ")) return false;
      if (!stralloc_cat(s, msg)) return false;
   }
   return true;
}

/* output `string' to `obuf' and check whether all bytes were accepted */
static bool outbuf_puts(outbuf* obuf, char* string) {
   int len = strlen(string);
   return outbuf_write(obuf, string, len) == len;
}

bool gen_request(outbuf* obuf, connect4_request* request) {
   char* cmdname = commands[request->command].name;
   return
      outbuf_puts(obuf, cmdname) &&
      (
	 request->param == 0? 1:
	    outbuf_putchar(obuf, ' ') == ' ' &&
	    outbuf_puts(obuf, request->param)
      ) &&
      outbuf_putchar(obuf, '\r') == '\r' &&
      outbuf_putchar(obuf, '\n') == '\n';
}

static bool expect_string(inbuf* ibuf, char* string) {
   char ch1;
   while ((ch1 = *string++)) {
      int ch2 = inbuf_getchar(ibuf);
      if (ch2 < 0 || ch1 != ch2) return false;
   }
   return true;
}

static bool parse_success(inbuf* ibuf, bool* success) {
   int ch = inbuf_getchar(ibuf);
   if (ch < 0) return false;
   if (ch == 'O' && expect_string(ibuf, "K")) {
      *success = true; return true;
   }
   if (ch == 'F' && expect_string(ibuf, "AILED")) {
      *success = false; return true;
   }
   return false;
}

bool parse_response(inbuf* ibuf, connect4_response* response) {
   bool success;
   if (!parse_success(ibuf, &success)) return false;
   int ch = inbuf_getchar(ibuf);
   if (ch < 0) return false;
   if (ch == ' ') {
      char* text = inbuf_readline(ibuf);
      if (!text) return false;
      int len = strlen(text);
      if (len > 0 && text[len-1] == '\r') {
	 text[len-1] = '\0'; /* remove CR */
      }
      response->succeeded = success;
      response->result = text;
      return true;
   } else if (ch == '\r' || ch == '\n') {
      if (ch == '\r') {
	 ch = inbuf_getchar(ibuf);
	 if (ch != '\n') return false;
      }
      response->succeeded = success;
      response->result = 0;
      return true;
   } else {
      return false;
   }
}
