#include <afblib/mpx_session.h>
#include <stralloc.h>
#include <afblib/strhash.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "connect4_packets.h"
#include "connect4.h"

enum user_state {IDLE, WAITING, PLAYING};
struct user {
   char* name; /* login name of this user, always != 0 */
   enum user_state state;
   session* s;
   struct game* game;
};
struct game {
   struct user* white;
   struct user* black;
   connect4_board board;
};

static strhash users = {0}; /* hash of (char* login, struct user* user) pairs */

/* return the player who has to move next */
struct user* next_player(struct game* g) {
   if (connect4_finished(&g->board)) return 0;
   switch (connect4_next(&g->board)) {
      case CONNECT4_BLACK:
	 return g->black;
      case CONNECT4_WHITE:
	 return g->white;
      default:
	 return 0;
   }
}

/* return the player participating in g who is not me */
struct user* opponent(struct game* g, struct user* me) {
   if (me == g->white) {
      return g->black;
   } else {
      return g->white;
   }
}

/* returns pointer to associated user structure, if existant, otherwise 0 */
struct user* get_user(char* login) {
   if (users.size == 0) return 0;
   char* cp;
   if (!strhash_lookup(&users, login, &cp)) return 0;
   return (struct user*) cp;
}

/* attempt to register a new user; return 0, if the name is already used */
struct user* register_user(char* login) {
   if (strhash_exists(&users, login)) return 0;
   struct user* user = (struct user*) malloc(sizeof(struct user));
   if (user == 0) return 0;
   *user = (struct user) {strdup(login)};
   strhash_add(&users, user->name, (char*) user);
   return user;
}

/* return an error response */
static void return_error(session* s, char* msg) {
   connect4_response response = {false, msg};
   gen_response(&response, &s->response);
   send_response(s);
}

/* return a general confirmative response */
static void return_ok(session* s, char* msg) {
   connect4_response response = {true, msg};
   gen_response(&response, &s->response);
   send_response(s);
}

/* return a random boolean value */
static bool flip() {
   return rand() % 2 == 0;
}

static void return_game_ok(struct user* u,
      char* state, connect4_move move, char* opponent) {
   stralloc msg = {0};
   stralloc_cats(&msg, state);
   if (move >= 1) {
      stralloc_cats(&msg, " ");
      stralloc_catint(&msg, move);
   }
   if (opponent) {
      stralloc_cats(&msg, " ");
      stralloc_cats(&msg, opponent);
   }
   gen_response_packet(true, &msg, &u->s->response);
   send_response(u->s);
   stralloc_free(&msg);
}

/* u waits for an invitation */
static void wait_for_invitation(struct user* u) {
   session* s = u->s;
   if (u->state == PLAYING) {
      return_error(s, "Please finish your currently running game first!");
      return;
   }
   if (u->state == WAITING) {
      return_error(s, "You are already waiting for an invitation!");
      return;
   }
   u->state = WAITING;
}

/* p1 attempts to invite p2 for a game */
static void start_game(session* s, struct user* p1, struct user* p2) {
   /* sanity checks */
   if (p1->state == PLAYING) {
      return_error(s, "Please finish your currently running game first!");
      return;
   }
   if (p1->state == WAITING) {
      return_error(s, "Please wait for someone else to invite you!");
      return;
   }
   if (p2->state != WAITING) {
      return_error(s, "The other user is currently not available!");
   }
   /* prepare everything for the new match */
   struct game* g = malloc(sizeof(struct game));
   if (g == 0) {
      return_error(s, "Out of memory!"); return;
   }
   if (flip()) {
      g->white = p1; g->black = p2;
   } else {
      g->white = p2; g->black = p1;
   }
   p1->state = PLAYING; p2->state = PLAYING;
   p1->game = g; p2->game = g;
   connect4_init(&g->board);
   /* let white move first */
   return_game_ok(g->white, "init", 0, g->black->name);
}

/* let both players drop back into the idle state
   and release all storage associated with g */
static void finish_game(struct game* g) {
   g->white->state = IDLE; g->black->state = IDLE;
   g->white->game = 0; g->black->game = 0;
   free(g);
}

/* u makes a move for a running match */
static void perform_move(struct user* u, char* move) {
   session* s = u->s;
   /* sanity checks */
   if (u->state != PLAYING) {
      return_error(s, "You are not yet playing!"); return;
   }
   if (move == 0) {
      return_error(s, "Parameter expected!"); return;
   }
   struct game* g = u->game;
   if (u != next_player(g)) {
      return_error(s, "It's not your turn!"); return;
   }
   int m = atoi(move);
   if (!connect4_valid_move(&g->board, m)) {
      return_error(s, "Invalid move!"); return;
   }
   /* move is valid, perform it and check if the game is finished */
   connect4_perform_move(&g->board, m);
   if (connect4_finished(&g->board)) {
      if (connect4_tied(&g->board)) {
	 return_game_ok(g->white, "tied", m, u->name);
	 return_game_ok(g->black, "tied", m, u->name);
      } else if (connect4_winner(&g->board) == CONNECT4_WHITE) {
	 return_game_ok(g->white, "won", m, u->name);
	 return_game_ok(g->black, "lost", m, u->name);
      } else {
	 return_game_ok(g->white, "lost", m, u->name);
	 return_game_ok(g->black, "won", m, u->name);
      }
      finish_game(g);
   } else {
      struct user* nextp = next_player(g);
      struct user* other = opponent(g, nextp);
      return_game_ok(nextp, "continue", m, other->name);
   }
}

/* list all users that are waiting for an invitation */
static void list_users(session* s) {
   stralloc userlist = {0};
   char* key;
   strhash_start(&users);
   while (strhash_next(&users, &key)) {
      char* cp;
      strhash_lookup(&users, key, &cp);
      struct user* u = (struct user*) cp;
      if (u->state != WAITING) continue;
      if (userlist.len > 0) {
	 stralloc_cats(&userlist, " ");
      }
      stralloc_cats(&userlist, key);
   }
   gen_response_packet(true, &userlist, &s->response);
   send_response(s);
   stralloc_free(&userlist);
}

/* request handler of mxp_session;
   note that responses are always sent by invoking send_response directly
*/
static void handle_request(session* s) {
   /* parse request */
   connect4_request request = {false};
   if (!parse_request(&s->request, &request)) {
      return_error(s, "Unknown command or invalid command syntax."); return;
   }
   /* permit always a session to be closed */
   if (request.command == CONNECT4_QUIT) {
      return_ok(s, "Bye!"); close_session(s); return;
   }
   /* take care of new sessions / logins */
   if (s->handle == 0) {
      if (request.command != CONNECT4_LOGIN) {
	 return_error(s, "Please login first."); return;
      }
      if (request.param == 0) {
	 return_error(s, "User name is missing."); return;
      }
      struct user* user = register_user(request.param);
      if (user == 0) {
	 return_error(s, "Someone else has already taken this name."); return;
      }
      user->s = s;
      s->handle = (void*) user;
      return_ok(s, "Welcome!"); return;
   } else if (request.command == CONNECT4_LOGIN) {
      return_error(s, "You are already logged in!"); return;
   }
   /* at this point we have an established session with an
      already registered user */
   struct user* user = (struct user*) s->handle;
   switch (request.command) {
      case CONNECT4_PLAY:
	 if (request.param == 0) {
	    wait_for_invitation(user);
	 } else {
	    struct user* opponent = get_user(request.param);
	    if (opponent == 0) {
	       return_error(s, "Unknown player!"); return;
	    }
	    start_game(s, user, opponent);
	 }
	 break;
      case CONNECT4_MOVE:
	 perform_move(user, request.param);
	 break;
      case CONNECT4_USERS:
	 list_users(s);
	 break;
      default:
	 return_error(s, "Not yet implemented!"); return;
   }
}

/* close handler of mpx_session */
void handle_close(session* s) {
   if (s->handle) {
      struct user* user = (struct user*) s->handle;
      if (user->state == PLAYING) {
	 /* notify opponent */
	 struct game* g = user->game;
	 struct user* opp = opponent(g, user);
	 return_error(opp->s, "Your opponent dropped out!");
	 finish_game(g);
      }
      strhash_remove(&users, user->name);
      free(user->name);
      free(s->handle);
   }
}

int main() {
   srand(getpid());
   strhash_alloc(&users, 32);
   run_mpx_service(CONNECT4_PORT, handle_request, handle_close);
}
